
-- ~~~~~~~~~~~~
-- Lander.lua
-- ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
-- Lander object for Mars Lander
-- ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
local Lander = {}

local gravity = 81



-- ~~~~~~~~~~~~~~~~
-- Local functions
-- ~~~~~~~~~~~~~~~~




-- ~~~~~~~~~~~~~~~~~
-- Public functions
-- ~~~~~~~~~~~~~~~~~

function Lander:init()
    self:setDimensions(16, 16)
    self:setPosition(screenWidth/2-self.width/2, 350)
    self.removed    = false
    self.angle      = 0
    self.vx         = 0
    self.vy         = 0
    self.rvel       = 0
    self.maxVel     = 300
    self.sideEngineThrust = 60
    self.mainEngineThrust = 130
    self.landed = false

    -- Colliders can be used to check points around the lander for collisions
    self.collider = {}


    -- The collision part is a mess! We'd need to redo it properly and more reliably and without hardcoded stuff

    -- LEFT LEG COLLISION
    local onCollision = function(collider, object)
        if self.vy < 70 and math.abs(self.angle) <= 15 then
            self.y = (object.y-collider.oy-collider.height/2)-collider.height/2
            if math.abs(self.angle) < 5 then
                self.landed = true
                return
            end
        elseif self.vy < 150 then
            self.y = (object.y-collider.oy-collider.height/2)-collider.height*3
            self.vy = -self.vy*.5
            self.vx = self.vx*.85
            self.rvel = -(self.vy*.3)+15
        else
            self:destroy()
        end
    end

    self:createCollider(-6, self.height+2, 3, 3, onCollision) -- left leg

    -- RIGHT LEG COLLISION
    local onCollision = function(collider, object)
        if self.vy < 70 and math.abs(self.angle) <= 12 then
            self.y = (object.y-collider.oy-collider.height/2)-collider.height/2
            if math.abs(self.angle) < 5 then
                self.landed = true
                return
            end
        elseif self.vy < 150 then
            self.y = (object.y-collider.oy-collider.height/2)-collider.height*3
            self.vy = -self.vy*.5
            self.vx = self.vx*.85
            self.rvel = (self.vy*.3)-15
        else
            self:destroy()
        end
    end

    self:createCollider(self.width+2, self.height+2, 3, 3, onCollision) -- right leg

    -- BODY COLLISION
    local onCollision = function(collider, object)
        self:destroy()
    end

    self:createCollider(2, 2, self.width-4, self.height-4, onCollision) -- body

    -- Initial position update
    self:updateCollider()
end



function Lander:setPosition(x, y)
    self.x = x or self.x or 0
    self.y = y or self.y or 0
end



function Lander:setDimensions(w, h)
    self.width = w or self.width or 32
    self.height = h or self.height or 32
end



function Lander:createCollider(ox, oy, w, h, onCollision)
    local collider = {}
    collider.width = w or 2
    collider.height = h or 2
    collider.ox = ox or 0
    collider.oy = oy or 0
    collider.x = self.x + collider.ox
    collider.y = self.y + collider.oy
    collider.onCollision = onCollision or function() end
    table.insert(self.collider, collider)
end



function Lander:updateCollider()
    for _, collider in ipairs(self.collider) do

        -- Calculate collider position based on offset values
        collider.x = self.x + collider.ox
        collider.y = self.y + collider.oy
        local center = {
            x = self.x + self.width / 2 - collider.width / 2,
            y = self.y + self.height / 2 - collider.height / 2,
        }
        -- Apply lander angle to collider position
        local angle_radian = math.rad(self.angle)
        local rotatedX = math.cos(angle_radian) * (collider.x-center.x) - math.sin(angle_radian) * (collider.y-center.y) + center.x;
        local rotatedY = math.sin(angle_radian) * (collider.x-center.x) + math.cos(angle_radian) * (collider.y-center.y) + center.y
        collider.x = rotatedX
        collider.y = rotatedY
    end
end



function Lander:checkCollisions(newx, newy)
    local groundBox = {x=0, y = screenHeight-128, width = screenWidth, height = 128}
    for _, collider in ipairs(self.collider) do
        if boundingBox(groundBox, collider) then
            collider:onCollision(groundBox)
        end
    end
end



function Lander:drawCollider()
    for _, collider in ipairs(self.collider) do
        love.graphics.rectangle("line", collider.x, collider.y, collider.width, collider.height)
    end
end



function Lander:move(dx, dy)
    local dt = love.timer.getDelta()
    self.x = self.x+dx*dt
    self.y = self.y+dy*dt
end



function Lander:accelerate(dx, dy)
    local dt = love.timer.getDelta()
    self.vx = self.vx+dx*dt
    self.vy = self.vy+dy*dt
end



function Lander:limitVelocity(x, y)
    self.vx = clamp(self.vx, -x, x)             -- x-axis
    self.vy = clamp(self.vy, -y, y)             -- y-axis
    self.rvel = clamp(self.rvel, -x, x)         -- rotation
end



function Lander:rotate(acceleration, dt)
    self.rvel = self.rvel+acceleration*dt
end



function Lander:destroy()
    self.removed = true
end


function Lander:update(dt)

    -- Calculate direction forces are applied to
    local angle_radian = math.rad(self.angle)
    local force_x = math.sin(-angle_radian)
    local force_y = math.cos(angle_radian)

    -- Main thruster (bottom)
    if love.keyboard.isDown("up") then
        self:accelerate(-(force_x*self.mainEngineThrust), -(force_y*self.mainEngineThrust))
        self.landed = false
    -- Control thruster (top)
    elseif love.keyboard.isDown("down") then
        self:accelerate(force_x*self.sideEngineThrust, force_y*self.sideEngineThrust)
    end
    
    -- Side thrusters (rotation)
    if love.keyboard.isDown("left") then self:rotate(-self.sideEngineThrust, dt)
    elseif love.keyboard.isDown("right") then self:rotate(self.sideEngineThrust, dt) end

    -- Simulating wind drag
    -- This could be reduced/removed when the player is flying high (less/no atmosphere)
    -- or influenced by parameters like x velocity, mass etc.
    if self.angle > 5 and self.vx > 120 then
        self:rotate(-5, dt)
    elseif self.angle < -5 and self.vx < -120 then
        self:rotate(5, dt)
    end

    -- Reset rotation value
    if math.abs(self.angle) >= 360 then self.angle = 0 end

    -- Gradually reduce x-axis velocity
    -- This could be reduced/removed when the player is flying high (less/no atmosphere)
    self.vx = self.vx*.998

    -- Gradually reduce rotation velocity
    -- This could be reduced/removed when the player is flying high (less/no atmosphere)
    self.rvel = self.rvel*.999

    -- Apply gravity effect
    self:accelerate(0, gravity)

    -- Apply velocities
    self:move(self.vx, self.vy)
    self.angle = self.angle+self.rvel*dt

    -- Clamp velocities to a max value
    self:limitVelocity(self.maxVel, self.maxVel)

    -- Limit screen bounds
    if self.x < 0 then
        self.x = 0
        self.vx = 0
    elseif self.x + self.width > screenWidth then
        self.x = screenWidth-self.width
        self.vx = 0
    end

    -- Update collider
    self:updateCollider(dt)

    self:checkCollisions()

    if self.landed and not love.keyboard.isDown("up") then
        self.vy = 0
        self.vx = self.vx*.98
        self.rvel = 0
        self.angle = 0
    end
end



function Lander:draw()
    love.graphics.push()
    -- Set angle center & rotate the image
    love.graphics.translate(self.x+self.width/2, self.y+self.height/2)
    love.graphics.rotate(math.rad(self.angle))
    love.graphics.translate(-self.x-self.width/2, -self.y-self.height/2)

    -- Draw the object
    love.graphics.rectangle("line", self.x, self.y, self.width, self.height)
    love.graphics.line(self.x + self.width/2, self.y, self.x + self.width/2, self.y + 8)
    love.graphics.points(self.x+self.width/2, self.y+self.height/2)             -- center of the object
    love.graphics.points(self.x-4, self.y+self.height+4)                        -- Left Leg Collision Point
    love.graphics.points(self.x+self.width+4, self.y+self.height+4)             -- Right Leg Collision Point
    love.graphics.pop()
    love.graphics.print("VX-SPEED: "..round(self.vx), 10, 10)
    love.graphics.print("VY-SPEED: "..round(self.vy), 10, 30)

    -- Debug draw the colliders
    self:drawCollider()
end


return Lander